###############=============================================================
# start fetchURL.tcl section
###############
#
#       Implements a fetchURL remote html document retrieval function from
#       atop the Tcl 7.5 socket command.  Binary transfers are supported
#       with the "-outfile" option.
#       options:
#          -url <url>                  -- url to retrieve
#          -outfile <filename>         -- place output in named file rather
#                                         than returning it from this 
#                                         procedure
#          -timeout <secs>             -- timeout between I/O buffers
#                                         before implying a connection 
#                                         closure (default 5 seconds)
#          -initialtimeout <secs>      -- timeout to wait for initial
#                                         connection establishment and
#                                         1st buffer retrieval (default
#                                         30 seconds)
#
#---------------------------------------------------------------------------

# -- validate appropriate execution environment
if {[catch "info tclversion" tclversion] || $tclversion < 7.5} {
    error "fetchURL.tcl requires Tcl7.5 / Tk4.1 or later releases"
} else {unset tclversion}
global sizevar
set sizevar ""
if {[catch "info tclversion" tclversion] || $tclversion > 7.6} {
set sizevar " -size "
}
# -- find and rename binary copy command
if {[info commands "unsupported0"] == "unsupported0"} {
    rename unsupported0 fcopy
}
if {[info commands "fcopy"] == ""} {
	error "fetchURL.tcl requires use of fcopy command - not found"
}

#----------
# -- set up for transferring url
proc Http_Copy {url {locfile ""} {chunk 4096}} {
    global http
    array set http {
        state      header
        mime       {}
        copyDone   0
        copySize   0
        document   {}
        all        {}
        protocol   {}
        host       {}
        p          {}
        port       {}
        what       {}
        key        {}
		ext        {}
        name       {}
        type       "text/html"
        length     0
        code       0
        location   {}
        timeoutid  {}
    }
	# -- parse the URL
	set http(full) $url
	set http(urllist) [parseurl $url]
	set http(protocol) [lindex $http(urllist) 0]
	set http(host) [lindex $http(urllist) 1]
	set http(port) [lindex $http(urllist) 2]
	set http(what) [lindex $http(urllist) 3]
	set http(ext) [lindex $http(urllist) 4]
	set http(tag) [lindex $http(urllist) 5]
	set http(key) [lindex $http(urllist) 6]
    if {"$http(port)" == ""} {set http(port) "80"}
    if {"$http(host)" == ""} {set http(host) [info hostname]}
    if {"$http(what)" == ""} {set http(what) "/"}
	# -- remove leading "/" if first this is dos drive specifier
	if {[string index $http(what) 0] == "/" && \
			[string index $http(what) 2] == ":" && \
			[string index $http(what) 3] == "/"} {
		set http(what) [string range $http(what) 1 end]
	}
    set out -1
    if {"$locfile" != ""} {
		if {[catch "open [list "$locfile"] w" out]} {
			puts stderr "fetchURL:  unable to open local destination file: $locfile"
			return 0
		}
        if {[catch "fconfigure $out -translation lf"]} {
			catch "close $out"
			puts stderr "fetchURL:  unable to configure local destination file descriptor"
			return 0
		}
    }
	# -- handle "file" protocol here
	if {$http(protocol) == "" || [string tolower $http(protocol)] == "file"} {
		if {(![file exists $http(what)]) || (![file readable $http(what)])} {
			set http(code) 404
			return [error404 $http(full)]
		}
		if {[file isdirectory $http(what)]} {
			if {"/" != [string index $http(what) [expr [string length $http(what)] - 1]]} {
				append http(what) "/"
				append http(full) "/"
			}
			append http(what) "index.html"
			if {(![file exists $http(what)]) || (![file readable $http(what)])} {
				set http(code) 404
				return [error404 "$http(full)index.html"]
			}
		}
		if {[catch "open [list "$http(what)"] r" in]} {
			set http(code) 403
			return [error403 $http(full)]
		}
		catch "fconfigure $in -translation lf"
		if {-1 != $out} {
			fcopy $in $out
			close $in
			close $out
			return 1
		} else {
			set buf "[read $in]"
			close $in
			return "$buf"
		}
	}
    set http(timeoutid) [after $http(inittimeoutlen) "HttpTimeout -1 $out"]
    set s [HttpOpen]
    if {"$s" == ""} {return 0}
	fileevent $s readable [list HttpCopy $s $out $chunk]
    vwait http(copyDone)
	catch "after cancel $http(timeoutid)"
	if {"$s" != -1} {
		catch "fileevent $s readable {}"
		catch "close $s"
	}
    if {"$out" != "-1"} {
        catch "close $out"
        return 1
    } else {
        return "$http(document)"
    }
}

#----------
# -- Open a connection to the http server
proc HttpOpen {} {
    global http
    set http(connectDone) 0
    set http(type) text/html
    if {[catch "socket $http(host) $http(port)" s]} {
        error "fetchURL:  could not open connection to $http(host) at port $http(port)"
    }
    if {[catch "fileevent $s writable [list "HttpConnect $s"]"]} {
		puts stderr "fetchURL:  unable to set up writable event handler"
		catch "fileevent $s writable {}"
		catch "close $s"
		return ""
	}
    vwait http(connectDone)
	if {[catch "eof $s" eofflag] || $eofflag} {
		puts stderr "fetchURL:  remote connection prematurely closed by server"
		catch "fileevent $s writable {}"
		catch "close $s"
		return ""
	}
	if {[catch "fconfigure $s -translation crlf"]} {
		puts stderr "fetchURL:  unable to set line protocol for MIME header"
		catch "fileevent $s writable {}"
		catch "close $s"
		return ""
	}
	set what $http(what)
	append what $http(key)
    catch "puts $s [list "GET $what HTTP/1.0"]"
    catch "puts $s [list "Accept: */*"]"
    catch "puts $s [list "User-Agent:  Spynergy Geturl"]"
    catch "puts $s {}"
    if {[catch "flush $s"]} {
		puts stderr "fetchURL:  unable to write MIME header for request"
		catch "fileevent $s writable {}"
		catch "close $s"
		return ""
	}
		
    # -- Put this back because the server may or may not use crlf 
    # -- in the headers
    if {[catch "fconfigure $s -translation auto"]} {
		puts stderr "fetchURL:  unable to set line protocol for document body"
		catch "fileevent $s writable {}"
		catch "close $s"
		return ""
	}
    return $s
}

#----------
# -- Callback made when connection is established
proc HttpConnect {s} {
    global http
    set http(connectDone) 1
    catch "fileevent $s writable {}"
}

#----------
# -- Procedure invoked when a timeout is hit
proc HttpTimeout {s outfd} {
    global http
    catch "after cancel $http(timeoutid)"
    set http(copyDone) 1
	if {$outfd != -1} {catch "close $outfd"}
}

#----------
# -- Process buffers received from server
proc HttpCopy {s outfd chunksize} {
    global http sizevar
    after cancel $http(timeoutid)
    set http(timeoutid) [after $http(timeoutlen) "HttpTimeout $s $outfd"]
    if {$http(state) == "header"} {
        if {[catch "gets $s line" n]} {set n -1}
		if {[catch "eof $s" eofflag] || $eofflag} {set n -1}
		if {$n < 0} {
            set http(copyDone) 1
        } elseif {$n == 0} {
            set http(state) "body"
            if {"$http(type)" == "text/html" || \
                    "$http(type)" == "application/tcl" || \
                    "$http(type)" == "application/x-spynergy" || \
                    "$http(type)" == "application/weblet"} {
                catch "fconfigure $s -translation auto"
            } else {
                catch "fconfigure $s -translation lf"
            }
            if {$http(code) == 301 || $http(code) == 302} {
                after cancel $http(timeoutid)
                set http(copyDone) 1
            }
        } else {
            # n > 0
            if {"content-type:" == "[lindex [string tolower $line] 0]"} {
                set http(type) "[lindex [string tolower $line] 1]"
            }
            if {"content-length:" == "[lindex [string tolower $line] 0]"} {
                set http(length) "[lindex [string tolower $line] 1]"
            }
            if {"http/1.0" == "[lindex [string tolower $line] 0]"} {
                set http(code) "[lindex [string tolower $line] 1]"
            }
            if {"location:" == "[lindex [string tolower $line] 0]"} {
                set http(location) "[lindex [string tolower $line] 1]"
            }
            append http(mime) "$line\n"
        }
    } else {
        if {$outfd != "-1"} {
			if {[catch "fcopy $s $outfd $sizevar $chunksize" n]} {set n -1}
        } else {
			if {[catch "read $s $chunksize" line]} {
				set n -1
			} else {
				set n [string length "$line"]
			}
            if {$n > 0} {append http(document) "$line"}
        }
		if {[catch "eof $s" eofflag] || $eofflag} {set n -1}
        if {$n <= 0} {
            set http(copyDone) 1
        } else {
            incr http(copySize) $n
            if {$http(length) != 0 && $http(copySize) >= $http(length)} {set http(copyDone) 1}
        }
    }
}

#----------
# -- wrapper procedure to get arguments and start things up
proc fetchURL {args} {
    global http
    set l [llength $args]
    set url ""
    set outfile ""
	set http(timeoutlen) 5000
	set http(inittimeoutlen) 30000
    for {set i 0} {$i < $l} {incr i} {
        set a [lindex $args $i]
        if {"[string tolower $a]" == "-url"} {
            incr i
            set url "[lindex $args $i]"
        } elseif {"[string tolower $a]" == "-outfile"} {
            incr i
            set outfile "[lindex $args $i]"
		} elseif {"[string tolower $a]" == "-timeout"} {
			incr i
			set timeout "[lindex $args $i]"
			if {[catch "expr int($timeout * 1000)" timeout]} {
				error "fetchURL: timeout value must be numeric"
			}
			set http(timeoutlen) $timeout
		} elseif {"[string tolower $a]" == "-initialtimeout"} {
			incr i
			set inittimeout "[lindex $args $i]"
			if {[catch "expr int($inittimeout * 1000)" inittimeout]} {
				error "fetchURL: inittimeout value must be numeric"
			}
			set http(inittimeoutlen) $inittimeout
        } else {
            set url "$a"
        }
    }
    set x "[Http_Copy $url $outfile]"
    if {$http(code) == 301 || $http(code) == 302} {
        if {$http(location) != ""} {
            if {$outfile == ""} {
                return "[fetchURL -url $http(location)]"
            } else {
                return "[fetchURL -url $http(location) -outfile $outfile]"
            }
        } else {
            return "$x"
        }
    } else {
        return "$x"
    }
}

#----------
# -- procedure to call if local file is not found
proc error404 {url} {
	set doc "<HEAD><TITLE>404 Not Found</TITLE></HEAD>
<BODY><H1>404 Not Found</H1>
The requested URL:  \"$url\" was not found.
</BODY>

"
    return $doc
}

#----------
# -- procedure to call if local file access had errors
proc error403 {url} {
	set doc "<HEAD><TITLE>403 Forbidden</TITLE></HEAD>
<BODY><H1>403 Forbidden</H1>
Obtaining the requested URL:  \"$url\" was denied.
</BODY>

"
    return $doc
}

#----------
# -- report background errors on the server
if {[info procs bgerror] == ""} {
    proc bgerror {args} {
        global errorInfo errorCode
        puts stderr "Background error: $args"
        puts stderr "\t$errorInfo"
        puts stderr "errorCode = \[$errorCode\]"
        return ""
    }
}

#============================================================================
# urlutil.tcl --
#        Procedures for parsing and manipulating URLs
#
# parseurl:
#        Breaks a URL into its components, returning these components in
#        a list:
#          index value:
#               0         protocol name (lowercase)
#               1         host name
#               2         port number
#               3         path
#               4         extension/MIME type ("." followed by lowercase)
#               5         tag found after "#"
#               6         key, including "?"
#
# mergeurls:
#         takes two urls and merges their paths.  The first argument
#         should be the current absolute path.  The second argument
#         can be any valid url or relative path.  The result is
#         an absolute url to the second argument.
#
# extracttagkey:
#         given an absolute URL, returns a list of three items:
#             absolute url without tag or key value
#             tag
#             key
#
#---------------------------------------------------------------------------

proc parseurl {url} {
    set protocol ""
    set host ""
    set port ""
    set path ""
    set ext ""
    set tag ""
    set key ""
    set full $url
    set i [string first "://" $url]
    if {$i != -1} {
        # -- parse out protocol
        set protocol [string tolower [string range $url 0 [expr $i - 1]]]
        set url [string range $url [expr $i + 3] end]
        # -- parse out host
        foreach c [split $url ""] {
            if {$c == "/" || $c == ":"} {break}
            append host $c
            set url [string range $url 1 end]
        }
        # -- parse out port number
        if {[string index $url 0] == ":"} {
            set url [string range $url 1 end]
            foreach c [split $url ""] {
                if {$c == "/"} {break}
                append port $c
                set url [string range $url 1 end]
            }
        }
    }
    # -- parse out path
    foreach c [split $url ""] {
        if {$c == "#" || $c == "?"} {break}
        append path $c
        set url [string range $url 1 end]
    }
    # -- parse out extension from path
    for {set i [expr [string length $path] - 1]} {$i>0} {incr i -1} {
        set c [string index $path $i]
        if {$c == "/"} {break}
        if {$c == "."} {
            set ext [string tolower [string range $path $i end]]
            break
        }
    }
    # -- parse out tag
    if {[string index $url 0] == "#"} {
        set url [string range $url 1 end]
        foreach c [split $url ""] {
            if {$c == "?"} {break}
            append tag $c
            set url [string range $url 1 end]
        }
    }
    # -- whatever remains must be a key
    set key $url
    return [list "$protocol" "$host" "$port" "$path" "$ext" "$tag" "$key"]
}

#----------
# -- merge urls into an absolute url
proc mergeurls {u1 u2} {
    set u1l [parseurl $u1]
    set u2l [parseurl $u2]
    set out ""
    # -- handle the case where u2 is absolute
    if {[lindex $u2l 0] != "" || [lindex $u2l 1] != "" || \
            [lindex $u2l 2] != ""} {
        if {[lindex $u2l 0] != ""} {
            append out "[lindex $u2l 0]://"
        }
        append out [lindex $u2l 1]
        if {[lindex $u2l 2] != ""} {
            append out ":[lindex $u2l 2]"
        }
        if {[string index [lindex $u2l 3] 0] != "/"} {
            append out "/"
        }
        append out [lindex $u2l 3]
        if {[lindex $u2l 5] != ""} {
            append out "#[lindex $u2l 5]"
        }
        if {[lindex $u2l 6] != ""} {
            append out [lindex $u2l 6]
        }
        return $out
    }
    # -- build URL prefix from the 1st URL
    if {[lindex $u1l 0] != ""} {
        append out "[lindex $u1l 0]://"
    }
    append out [lindex $u1l 1]
    if {[lindex $u1l 2] != ""} {
        append out ":[lindex $u1l 2]"
    }
    # -- merge the paths
    set p1 [lindex $u1l 3]
    set p2 [lindex $u2l 3]
	if {[string index $u2 0] == "#"} {
		# -- handle special case that relative link is only a tag
		append out [lindex $u1l 3]
		append out "#[lindex $u2l 5]"
		if {[lindex $u2l 6] != ""} {
			append out [lindex $u2l 6]
		}
		return $out
	}
    if {[string index $p1 [expr [string length $p1] - 1]] == "/"} {
        append p1 "xyz"
    }
    if {[string index $p2 0] == "/"} {
        set p1 $p2
    } else {
        set outp ""
        set p1 [split [file dirname $p1] "/"]
        set p1last [expr [llength $p1] - 1]
        set p2 [split $p2 "/"]
        set p2last [expr [llength $p2] - 1]
        foreach item $p2 {
            switch -exact -- $item {
                .. {
                    incr p1last -1
                    if {$p1last < -1} {
                        set p1last -1
                        set p1 ""
                    } else {
                        set p1 [lrange $p1 0 $p1last]
                    }
                }
                . {
                }
                default {
                    incr p1last
                    lappend p1 $item
                }
            }
        }
        if {[llength $p1] > 0 && [lindex $p1 0] != ""} {
            set p1 "{} $p1"
        }
        set p1 [join $p1 "/"]
        regsub -all {//} $p1 "/" p1
        regsub -all {//} $p1 "/" p1
        if {$p1 == ""} {set p1 "/"}
    }
    append out $p1
    if {[lindex $u2l 5] != ""} {
        append out "#[lindex $u2l 5]"
    }
    if {[lindex $u2l 6] != ""} {
        append out [lindex $u2l 6]
    }
    return $out
}

#----------
# -- extract tag and key from absolute url
proc extracttagkey {url} {
	set l [parseurl $url]
	if {[lindex $l 5] == "" && [lindex $l 6] == ""} {
		return [list "$url" "" ""]
	}
	set out ""
    if {[lindex $l 0] != ""} {
        append out "[lindex $l 0]://"
    }
    append out [lindex $l 1]
    if {[lindex $l 2] != ""} {
        append out ":[lindex $l 2]"
    }
	append out [lindex $l 3]
	return [list "$out" "[lindex $l 5]" "[lindex $l 6]"]
}


###############
# end fetchURL.tcl section
###############


